//
//  Z80.java
//  JuiceZ
//
//  Created by Thomas Dickerson on 5/5/09.
//  Copyright 2009 __MyCompanyName__. All rights reserved.
//

package sfgp.emu;
import sfgp.debug.Debugger;

public class Z80 {
	
	public byte r;
	public byte i;
	public short af, bc, de, hl, ix, iy, sp, pc;
	public short af_shadow, bc_shadow, de_shadow, hl_shadow;
	public byte[] memory;
	public boolean iff1, iff2;
	
	//emulation specific
	public boolean limitcpu;
	public int stepsize;
	public int stepper;
	
	private static byte boolbyte(boolean b){ return (byte)(b ? 1 : 0);	}
	private static char memval(short value){ return (char)(value); }
	private static void dieIllegalBitfield(){
		Debugger.printError("Illegal Bitfield");
		System.exit(1);
	}
	
	public Z80(int clock, boolean l){
		memory = new byte[65536];
		af = bc = de = hl = (short)0x0000;
		ix = iy = (short)0x0000;
		af_shadow = bc_shadow = de_shadow =hl_shadow = (short)0x0000;
		sp = (short)0xf000;
		pc = (short)0x0000;
		i = (byte)0x00;
		r = (byte)0x00;
		iff1 = false;
		iff2 = false;
		
		limitcpu = l;
		stepper = 0;
		stepsize = (clock - (clock % 50)) / 50;
	}
	
	private void cycles(int count){	stepper += count;	}
	
	public void setFlags(byte s, byte z, byte undoc5, byte h, byte undoc3, byte pv, byte n, byte c){
		byte flags = Util.getLowerEight(af);
		if (s >= 0) flags = (s==1) ? Util.setNthBit(flags, (byte)7) : Util.resetNthBit(flags, (byte)7);
		if (z >= 0) flags = (z==1) ? Util.setNthBit(flags, (byte)6) : Util.resetNthBit(flags, (byte)6);
		if (undoc5 >= 0) flags = (undoc5==1) ? Util.setNthBit(flags, (byte)5) : Util.resetNthBit(flags, (byte)5);
		if (h >= 0) flags = (h==1) ? Util.setNthBit(flags, (byte)4) : Util.resetNthBit(flags, (byte)4);
		if (undoc3 >= 0) flags = (undoc3==1) ? Util.setNthBit(flags, (byte)3) : Util.resetNthBit(flags, (byte)3);
		if (pv >= 0) flags = (pv==1) ? Util.setNthBit(flags, (byte)2) : Util.resetNthBit(flags, (byte)2);
		if (n >= 0) flags = (n==1) ? Util.setNthBit(flags, (byte)1) : Util.resetNthBit(flags, (byte)1);
		if (c >= 0) flags = (c==1) ? Util.setNthBit(flags, (byte)0) : Util.resetNthBit(flags, (byte)0);
		af = Util.setLowerEight(af, flags);
	}
	
	public byte getIReg16(boolean flag){	return memory[memval(flag ? de : bc)];	}
	public void setIReg16(byte val, boolean flag){	memory[memval(flag ? de : bc)] = val;	}
	
	public short getBitfieldIXOrIY(byte bitfield){ return ((bitfield & 0x20) != 0) ? iy : ix; }
	public void setBitfieldIXOrIY(short val, byte bitfield){
		if((bitfield & 0x20) == 0) ix = val;
		else iy = val;
	}
	
	public void setReg8Bitfield(byte val, byte bitfield){
		switch(bitfield){
			case 0:
				bc = Util.setUpperEight(bc, val);
				break;
			case 1:
				bc = Util.setLowerEight(bc, val);
				break;
			case 2:
				de = Util.setUpperEight(de, val);
				break;
			case 3:
				de = Util.setLowerEight(de, val);
				break;
			case 4:
				hl = Util.setUpperEight(hl, val);
				break;
			case 5:
				hl = Util.setLowerEight(hl, val);
				break;
			case 6:
				memory[memval(hl)] = val;
				break;
			case 7:
				af = Util.setUpperEight(af, val);
				break;
			default:
				dieIllegalBitfield();
		}
	}
	
	public byte getReg8Bitfield(byte bitfield){
		switch(bitfield){
			case 0:
				return Util.getUpperEight(bc);
			case 1:
				return Util.getLowerEight(bc);
			case 2:
				return Util.getUpperEight(de);
			case 3:
				return Util.getLowerEight(de);
			case 4:
				return Util.getUpperEight(hl);
			case 5:
				return Util.getLowerEight(hl);
			case 6:
				return memory[memval(hl)];
			case 7:
				return Util.getUpperEight(af);
			default:
				dieIllegalBitfield();
				return 0;
		}
	}
	
	public void setReg16Bitfield(short val, byte bitfield, boolean stackop){
		if(bitfield == 0) bc = val;
		else if(bitfield == 1) de = val;
		else if(bitfield == 2) hl = val;
		else if(bitfield == 3 && stackop) af = val;
		else if(bitfield == 3 && !stackop) sp = val;
		else dieIllegalBitfield();
	}
	
	public short getReg16Bitfield(byte bitfield, boolean stackop){
		if(bitfield == 0) return bc;
		else if(bitfield == 1) return de;
		else if(bitfield == 2) return hl;
		else if(bitfield == 3 && stackop) return af;
		else if(bitfield == 3 && !stackop) return sp;
		else {
			dieIllegalBitfield();
			return 0;
		}
	}
	
	public byte getIndexOfs(byte bitfield, byte ofs){	return memory[memval((short)(getBitfieldIXOrIY(bitfield) + ofs))];	}
	public void setIndexOfs(byte val, byte bitfield, byte ofs){	memory[memval((short)(getBitfieldIXOrIY(bitfield) + ofs))] = val;	}
	
	public void set_a(byte val){	af = Util.setUpperEight(af, val);	}
	public byte get_a(){	return Util.getUpperEight(af);	}
	
	public void addCarry(byte b){
		byte a = get_a();
		short c = (short)(a + b + Util.getNthBit(Util.getLowerEight(af), (byte)(0)));
		byte ans = (byte)c;
		setFlags(Util.getNthBit(ans, (byte)(7)), boolbyte(ans == 0), Util.getNthBit(ans, (byte)(5)), Util.getNthBit(ans, (byte)(3)), Util.getNthBit(ans, (byte)(3)), boolbyte((a & 0x7f) != (ans & 0x7f)), (byte)(0), boolbyte(c != ans));
		set_a(ans);
	}
	
	public void addCarry(short sh){
		int c = hl + sh + Util.getNthBit(Util.getLowerEight(af), (byte)(0));
		short ans = (short)c;
		byte upper = Util.getUpperEight(ans);
		setFlags(Util.getNthBit(upper, (byte)(7)), boolbyte(ans == 0), Util.getNthBit(upper, (byte)(5)), (byte)(-1), Util.getNthBit(upper, (byte)(3)), boolbyte((upper & 0x7f) != (Util.getUpperEight(hl) & 0x7f)), (byte)(0), boolbyte(c != ans));
		hl = ans;
	}
	
	
	
	public void ex_dehl(){
		cycles(4);
		short tmp = de;
		de = hl;
		hl = tmp;
	}
	
	public void ex_afaf_shadow(){
		cycles(4);
		short tmp = af;
		af = af_shadow;
		af_shadow = af;
	}
	
	public void ex_isphl(){
		cycles(19);
		byte low = memory[memval(sp)];
		byte high = memory[memval((short)(sp+1))];
		memory[memval(sp)] = Util.getLowerEight(hl);
		memory[memval((short)(sp+1))] = Util.getUpperEight(hl);
		hl = Util.twoBytesToWord(high, low);
	}
	
	public void ex_ispindex(byte bitfield){
		cycles(23);
		byte low = memory[memval(sp)];
		byte high = memory[memval((short)(sp+1))];
		short tmp = getBitfieldIXOrIY(bitfield);
		memory[memval(sp)] = Util.getLowerEight(tmp);
		memory[memval((short)(sp+1))] = Util.getUpperEight(tmp);
		setBitfieldIXOrIY(Util.twoBytesToWord(high, low), bitfield);
	}
	
	public void exx(){
		cycles(4);
		short tmpbc = bc;
		short tmpde = de;
		short tmphl = hl;
		bc = bc_shadow;
		de = de_shadow;
		hl = hl_shadow;
		bc_shadow = tmpbc;
		de_shadow = tmpde;
		hl_shadow = tmphl;
	}
	
	public void ld_reg8reg8(byte regd, byte regs){
		cycles((regd == 6 || regs == 6) ? 7 : 4);
		setReg8Bitfield(regd, getReg8Bitfield(regs));
	}
	
	public void ld_reg8imm8(byte regd, byte imm){
		cycles((regd == 6) ? 10 : 7);
		setReg8Bitfield(imm, regd);
	}
	
	public void ld_reg8ofsindex(byte regd, byte index, byte ofs){
		cycles(19);
		setReg8Bitfield(getIndexOfs(index, ofs), regd);
	}
	
	public void ld_ofsindexreg8(byte index, byte ofs, byte regs){
		cycles(19);
		setIndexOfs(getReg8Bitfield(regs), index, ofs);
	}
	
	public void ld_ofsindeximm8(byte index, byte ofs, byte val){
		cycles(19);
		setIndexOfs(val, index, ofs);
	}
	
	public void ld_aireg16(boolean flag){
		cycles(7);
		set_a(getIReg16(flag));
	}
	
	public void ld_aiimm16(short imm){
		cycles(13);
		set_a(memory[memval(imm)]);
	}
	
	public void ld_ireg16a(boolean flag){
		cycles(7);
		setIReg16(get_a(), flag);
	}
	
	public void ld_iimm16a(short imm){
		cycles(13);
		memory[memval(imm)] = get_a();
	}
	
	public void ld_air(boolean flag){
		cycles(9);
		set_a(flag ? r : i);
		byte a = get_a();
		setFlags(Util.getNthBit(a, (byte)7), boolbyte(a==0), Util.getNthBit(a, (byte)5), (byte)0, Util.getNthBit(a, (byte)3), (byte)0, boolbyte(iff2), (byte)(-1));
	}
	
	public void ld_ira(boolean flag){
		cycles(9);
		if (flag) r = get_a();
		else i = get_a();
	}
	
	public void ld_reg16imm16(byte regd, short imm){
		cycles(10);
		setReg16Bitfield(imm, regd, false);
	}
	
	public void ld_indeximm16(byte index, short imm){
		cycles(14);
		setBitfieldIXOrIY(imm, index);
	}
	
	public void ld_hliimm16(short imm){
		cycles(16);
		hl = Util.twoBytesToWord(memory[memval((short)(imm+1))], memory[memval(imm)]);
	}
	
	public void ld_reg16iimm16(byte regd, short imm){
		cycles(20);
		setReg16Bitfield(Util.twoBytesToWord(memory[memval((short)(imm+1))], memory[memval(imm)]), regd, false);
	}
	
	public void ld_indexiimm16(byte index, short imm){
		cycles(20);
		setBitfieldIXOrIY(Util.twoBytesToWord(memory[memval((short)(imm+1))], memory[memval(imm)]), index);
	}
	
	
	public void ld_iimm16hl(short imm){
		cycles(16);
		memory[memval(imm)] = Util.getLowerEight(hl);
		memory[memval((short)(imm+1))] = Util.getUpperEight(hl);
	}
	
	public void ld_iimm16reg16(short imm, byte regs){
		cycles(20);
		short tmp = getReg16Bitfield(regs, false);
		memory[memval(imm)] = Util.getLowerEight(tmp);
		memory[memval((short)(imm+1))] = Util.getUpperEight(tmp);
		
	}
	
	public void ld_iimm16index(short imm, byte index){
		cycles(20);
		short tmp = getBitfieldIXOrIY(index);
		memory[memval(imm)] = Util.getLowerEight(tmp);
		memory[memval((short)(imm+1))] = Util.getUpperEight(tmp);
		
	}
	
	public void ld_sphl(){
		cycles(6);
		sp = hl;
	}
	
	public void ld_spindex(byte index){
		cycles(10);
		sp = getBitfieldIXOrIY(index);
	}
	
	public void ldd(){
		cycles(16);
		memory[memval(de)] = memory[memval(hl)];
		byte tbita = (byte)(memory[memval(de)] + get_a());
		de--;
		hl--;
		bc--;
		if (bc != 0) setFlags((byte)(-1), (byte)(-1), Util.getNthBit(tbita, (byte)(5)), (byte)(0), Util.getNthBit(tbita, (byte)(3)), (byte)(1), (byte)(0), (byte)(-1));
		else setFlags((byte)(-1), (byte)(-1), Util.getNthBit(tbita, (byte)(5)), (byte)(0), Util.getNthBit(tbita, (byte)(3)), (byte)(0), (byte)(0), (byte)(-1));
	}
	
	public void lddr(){
		cycles(16);
		ldd();
		if (Util.getNthBit(Util.getLowerEight(af), (byte)(2)) != 0){
			cycles(5);
			pc -= 2;
		}
	}
	
	public void ldi(){
		cycles(16);
		// these instructions are weird. idk why.
		memory[memval(de)] = memory[memval(hl)];
		byte tbita = (byte)(memory[memval(de)] + get_a());
		de++;
		hl++;
		bc--;
		if (bc != 0) setFlags((byte)(-1), (byte)(-1), Util.getNthBit(tbita, (byte)(5)), (byte)(0), Util.getNthBit(tbita, (byte)(3)), (byte)(1), (byte)(0), (byte)(-1));
		else setFlags((byte)(-1), (byte)(-1), Util.getNthBit(tbita, (byte)(5)), (byte)(0), Util.getNthBit(tbita, (byte)(3)), (byte)(0), (byte)(0), (byte)(-1));
	}
	
	public void ldir(){
		cycles(16);
		ldi();
		if (Util.getNthBit(Util.getLowerEight(af), (byte)(2)) != 0){
			cycles(5);
			pc -= 2;
		}
	}
	
	public void pop_reg16(byte regd){
		cycles(10);
		short tmp = Util.twoBytesToWord(memory[memval((short)(sp+1))], memory[memval(sp)]);
		sp += 2;
		setReg16Bitfield(tmp, regd, true);
	}
	
	public void pop_index(byte index){
		cycles(14);
		short tmp = Util.twoBytesToWord(memory[memval((short)(sp+1))], memory[memval(sp)]);
		sp += 2;
		setBitfieldIXOrIY(tmp, index);
	}
	
	public void push_reg16(byte regs){
		cycles(11);
		short val = getReg16Bitfield(regs, true);
		sp--;
		memory[memval(sp)] = Util.getUpperEight(val);
		sp--;
		memory[memval(sp)] = Util.getLowerEight(val);
	}
	
	public void push_index(byte index){
		cycles(11);
		short val = getBitfieldIXOrIY(index);
		sp--;
		memory[memval(sp)] = Util.getUpperEight(val);
		sp--;
		memory[memval(sp)] = Util.getLowerEight(val);
	}
	
	public void adc_areg8(byte bitfield){
		cycles((bitfield == 0x06) ? 7 : 4);
		addCarry(getReg8Bitfield(bitfield));	
	}
	
	public void adc_aimm8(byte num){	
		cycles(7);
		addCarry(num);
	}
	
	public void adc_aofsindex(byte index, byte ofs){
		cycles(19);
		addCarry(getIndexOfs(index, ofs));
	}
	
	public void adc_hlreg16(byte bitfield){
		cycles(15);
		addCarry(getReg16Bitfield(bitfield, false));
	}
	
	
	
	
	
	public void nop(){	cycles(4);	}
	
	public void incr(){	r = (byte)(((r + 1) & 0x7f) | (r & 0x80));	}
	public byte fetchInstr(byte ofs){
		incr();
		return memory[memval((short)(pc + ofs))];
	}
	
	public int execNSteps(int n){
		byte op, op2;
		stepper = 0;
		while(stepper < n){
			op = fetchInstr((byte)(0));
			if (op == 0xdd || op ==0xfd){
				op2 = fetchInstr((byte)(1));
				if (op2 == 0xe3){
					ex_ispindex(op);
					pc += 2;
				} else if(op2 == 0x36){
					ld_ofsindeximm8(op, memory[memval((short)(pc+2))], memory[memval((short)(pc+3))]);
					pc += 4;
				} else if(op2 == 0x21){
					ld_indeximm16(op, Util.twoBytesToWord(memory[memval((short)(pc+3))], memory[memval((short)(pc+2))]));
					pc += 4;
				} else if(op2 == 0x2a){
					ld_indexiimm16(op, Util.twoBytesToWord(memory[memval((short)(pc+3))], memory[memval((short)(pc+2))]));
					pc += 4;
				} else if((op2 & 0xc7) == 0x46){
					byte regd = (byte)((op2 & 0x38) >>> 3);
					ld_reg8ofsindex(regd, op, memory[memval((short)(pc+2))]);
					pc += 3;
				} else if((op2 & 0xf8) == 0x70){
					byte regs = (byte)(op & 0x07);
					ld_ofsindexreg8(op, memory[memval((short)(pc+2))], regs);
					pc += 3;
				} else if(op2 == 0x22){
					ld_iimm16index(Util.twoBytesToWord(memory[memval((short)(pc+3))], memory[memval((short)(pc+2))]), op);
					pc += 4;
				} else if( op2 == 0xf9){
					ld_spindex(op);
					pc += 2;
				} else if( op2 == 0xe1){
					pop_index(op);
					pc += 2;
				} else if ( op2 == 0xe5){
					push_index(op);
					pc += 2;
				}
			} else if( op == 0xeb){
				ex_dehl();
				pc += 1;
			} else if( op == 0x08){
				ex_afaf_shadow();
				pc += 1;
			} else if( op == 0xe3){
				ex_isphl();
				pc += 1;		
			} else if( op == 0xd9){
				exx();
				pc += 1;
			} else if( (op & 0xc0) == 0x40){
				byte regs = (byte)(op & 0x07);
				byte regd = (byte)((op & 0x38) >>> 3);
				ld_reg8reg8(regd, regs);
				pc += 1;
			} else if( (op & 0xc7) == 0x06){
				byte regd = (byte)((op & 0x38) >>> 3);
				ld_reg8imm8(regd, memory[memval((short)(pc+1))]);
				pc += 2;
			} else if( (op & 0xef) == 0x0a){
				ld_aireg16(((op & 0x10) >>> 4) != 0);
				pc += 1;
			} else if( op == 0x3a){
				ld_aiimm16(Util.twoBytesToWord(memory[memval((short)(pc+2))], memory[memval((short)(pc+1))]));
				pc += 3;
			} else if( (op & 0xef) == 0x02){
				ld_ireg16a(((op & 0x10) >>> 4) != 0);
				pc += 1;
			} else if( op == 0x32){
				ld_iimm16a(Util.twoBytesToWord(memory[memval((short)(pc+2))], memory[memval((short)(pc+1))]));
				pc += 3;
			} else if( op == 0xed){
				op2 = fetchInstr((byte)1);
				if( (op2 & 0xf7) == 0x57){
					ld_air(((op2 & 0x08) >>> 3) != 0);
					pc += 2;
				} else if( (op2 & 0xf7) == 0x47){
					ld_ira(((op2 & 0x08) >>> 3) != 0);
					pc += 2;
				} else if( (op2 & 0xcf) == 0x4b){
					ld_reg16iimm16((byte)((op2 & 0x30) >>> 4), Util.twoBytesToWord(memory[memval((short)(pc+3))], memory[memval((short)(pc+2))]));
					pc += 4;
				} else if ( (op2 & 0xcf) == 0x43){
					ld_iimm16reg16(Util.twoBytesToWord(memory[memval((short)(pc+3))], memory[memval((short)(pc+2))]), (byte)((op2 & 0x30) >>> 4));
					pc += 4;
				} else if ( op2 == 0xa8){
					ldd();
					pc += 2;
				} else if ( op2 == 0xb8){
					lddr();
					pc += 2;
				} else if ( op2 == 0xa0){
					ldi();
					pc += 2;
				} else if ( op2 == 0xb0){
					ldir();
					pc += 2;
				}
						
			} else if( (op & 0xcf) == 0x01){
				ld_reg16imm16((byte)((op & 0x30) >>> 4), Util.twoBytesToWord(memory[memval((short)(pc+2))], memory[memval((short)(pc+1))]));
				pc += 3;
			} else if( op == 0x2a){
				ld_hliimm16(Util.twoBytesToWord(memory[memval((short)(pc+2))], memory[memval((short)(pc+1))]));
				pc += 3;
			} else if( op == 0x22){
				ld_iimm16hl(Util.twoBytesToWord(memory[memval((short)(pc+2))], memory[memval((short)(pc+1))]));
				pc += 3;
			} else if( op == 0xf9){
				ld_sphl();
				pc++;
			} else if ((op & 0xcf) == 0xc1){
				pop_reg16((byte)((op & 0x30) >>> 4));
				pc++;
			} else if ((op & 0xcf) == 0xc5){
				push_reg16((byte)((op & 0x30) >>> 4));
				pc++;
			} else{
				nop();
				pc++;	
			}
		}
		return stepper;
	}
	
	public void exec(){
		long start = System.currentTimeMillis();
		long now = System.currentTimeMillis();
		long cycles = 0;
		while (now - start < 4000){
			long zstart = System.currentTimeMillis();
			cycles += execNSteps(stepsize);
			now = System.currentTimeMillis();
			double pct = ((now - zstart <= 0) ? 1 : (50 / (double)(now - zstart))) * (stepper / (double)stepsize);
			if ((pct - 1) > 0.05 && limitcpu){
				try{					
					Thread.sleep((long)(1000 * (0.02 - (0.02/pct))));
				} catch (InterruptedException e){
					Debugger.printError("interrupting cow");
				}
				Debugger.printChatter("Running at 100% percent");
			} else{
				Debugger.printChatter("Running at " + (int)(100*pct) + " percent");
				try{					
					Thread.sleep(1l);
				} catch (InterruptedException e){
					Debugger.printError("interrupting cow");
				}
			}
			}
		Debugger.printInfo("" + cycles + " cycles complete in 4s");
	}

}
